/*
        This file is part of the cFos project,
        Copyright (c) 1993 by Martin Winkler, Christoph Lders

        This file contains low-level CAPI call routines used in capi.c
        and cstates.c.


        You are allowed to use, copy, distribute, publish, redistribute
        and/or modify this program under the terms of the 'cFos public
        licence and warranty'.

        This program is distributed in the hope that it will be useful,
        but WITHOUT ANY WARRANTY; without even the implied warranty of
        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

        See COPYING.CF for the 'cFos public license and warranty'.
*/


#include <string.h>

#include "vars.h"

char *itoa (int, char*, int);
static char buff[200];


#ifndef NDEBUG
static char *action_text (short cmd, short scmd)
{
   static char text[50];

   switch (cmd) {
      case 0x00: strcpy (text, "INIT"); break;
      case 0x01: strcpy (text, "RESET"); break;
      case 0x02: strcpy (text, "CONNECT"); break;
      case 0x03: strcpy (text, "CONNECT_ACTIVE"); break;
      case 0x04: strcpy (text, "DISCONNECT"); break;
      case 0x05: strcpy (text, "LISTEN"); break;
      case 0x06: strcpy (text, "GET_PARAMS"); break;
      case 0x07: strcpy (text, "INFO"); break;
      case 0x08: strcpy (text, "DATA"); break;
      case 0x09: strcpy (text, "CONNECT_INFO"); break;
      case 0x40: strcpy (text, "B2_PROTOCOL"); break;
      case 0x80: strcpy (text, "B3_PROTOCOL"); break;
      case 0x81: strcpy (text, "LISTEN_B3"); break;
      case 0x82: strcpy (text, "CONNECT_B3"); break;
      case 0x83: strcpy (text, "CONNECT_B3_ACT"); break;
      case 0x84: strcpy (text, "DISCONNECT_B3"); break;
      case 0x85: strcpy (text, "GET_B3_PARAMS"); break;
      case 0x86: strcpy (text, "DATA_B3"); break;
      case 0x87: strcpy (text, "HANDSET"); break;
      case 0xFF: strcpy (text, "MANUFACTURER"); break;
      default: strcpy (text, "?"); break;
   }
   switch (scmd) {
      case 0x00: strcat (text, "_REQ"); break;
      case 0x01: strcat (text, "_CONF"); break;
      case 0x02: strcat (text, "_IND"); break;
      case 0x03: strcat (text, "_RESP"); break;
      default: strcat (text, "_?"); break;
   }

   return text;
}
#endif


int api_installed (void)
{
   char _FAR * _FAR *api_isr_ptr;

   api_isr_ptr = (char _FAR * _FAR *) (dword) (capi_int << 2);
   return (*api_isr_ptr) [9] == 'I' && (*api_isr_ptr) [10] == 'A';
}


int api_register (void)
{
   union REGS r;
   struct SREGS s;

   mysegread (&s);
   r.h.ah = 0x01;                       /* API_REGISTER */
   r.x.cx = capi_maxmess;
   r.x.dx = capi_maxb3;
   r.x.si = capi_maxwin;
   r.x.di = capi_framelen;
   r.x.bx = FP_OFF (api_memptr);
   s.es = FP_SEG (api_memptr);
   myint86x (capi_int, &r, &r, &s);

   applid = r.x.ax;

   return applid == 0? (r.x.bx == 0? 0xffff: r.x.bx): 0;
}


word api_release (void)
{
   union REGS r;

   r.h.ah = 0x02;
   r.x.dx = applid;
   myint86 (capi_int, &r, &r);

   applid = 0;

   return r.x.ax;
}


void api_putmsg (void)
{
   union REGS r;
   struct SREGS s;
#ifndef NDEBUG
   word old_flags = get_flags ();
#endif

#ifndef NDEBUG
   adisp (action_text (smsgbuf[in_event][4], smsgbuf[in_event][5]));
   _adisp_char ('\n');
#endif

   mysegread (&s);
   r.h.ah = 0x03;
   s.es = FP_SEG (smsgbuf[in_event]);
   r.x.bx = FP_OFF (smsgbuf[in_event]);
   r.x.dx = applid;
   myint86x (capi_int, &r, &r, &s);
#ifndef NDEBUG
   if ((old_flags & 0x200) != (get_flags () & 0x200))
      myaprintf ("api_putmsg: changed I-flag\n");
#endif

#ifndef NDEBUG
   if (r.x.ax != 0) {
      myaprintf ("api_putmsg: cause=%s\n", capi_cause (r.x.ax));
      dump (smsgbuf[in_event], *(word*) (smsgbuf[in_event]));
   }
#endif
   smsgnr = (smsgnr+1) & 0x7fff;

   /* return r.x.ax; */
}


word api_getmsg (void)
{
   union REGS r;
   struct SREGS s;
#ifndef NDEBUG
   word old_flags = get_flags ();
#endif

   mysegread (&s);
   r.h.ah = 0x04;
   r.x.dx = applid;
   myint86x (capi_int, &r, &r, &s);
#ifndef NDEBUG
   if ((old_flags & 0x200) != (get_flags () & 0x200))
      myaprintf ("api_getmsg: changed I-flag\n");
#endif

#ifndef NDEBUG
   if (r.x.ax != 0 && r.x.ax != 0x1006) {
      myaprintf ("api_getmsg: cause=%s\n", capi_cause (r.x.ax));
      dump (msgbuf, *(word*) msgbuf);
   }
#endif

   if (r.x.ax == 0x1007)
      r.x.ax = 0;
   memcpy3 (msgbuf, MK_FP (s.es, r.x.bx), MSGLEN);

#ifndef NDEBUG
   if (r.x.ax == 0) {
      adisp (action_text (msgbuf[4], msgbuf[5]));
      _adisp_char ('\n');
   }
#endif

   return r.x.ax;
}


word api_set_signal (void _FAR *ptr)
{
   union REGS r;
   struct SREGS s;

   mysegread (&s);
   r.h.ah = 0x05;
   r.x.dx = applid;
   r.x.bx = FP_OFF (ptr);
   s.es = FP_SEG (ptr);
   myint86x (capi_int, &r, &r, &s);

#ifndef NDEBUG
   if (r.x.ax != 0)
      myaprintf ("api_set_signal: cause=%s\n", capi_cause (r.x.ax));
#endif

   return r.x.ax;
}


char *capi_cause (word number)
{
   char *temp;

   buff [mysprintf (buff, "%#04x: ", number)] = 0;

#ifndef NDEBUG
   switch (number) {
      case 0x0000: temp="Kein Fehler"; break;
      case 0x1001: temp="Fehler beim Registrieren";break;
      case 0x1002: temp="Applikationsnummer falsch";break;
      case 0x1003: temp="Message zu klein oder Messagenummer falsch kodiert";break;
      case 0x1004: temp="Falsches Command oder Subcommand";break;
      case 0x1005: temp="Message-Queue voll";break;
      case 0x1006: temp="Message-Queue war leer";break;
      case 0x1007: temp="Min. eine Message ist verlorengegangen, da Message-Queue voll";break;
      case 0x1008: temp="Fehler bei Deinstallation";break;

      case 0x2001: temp="Falscher Controller";break;
      case 0x2002: temp="Falscher PLCI";break;
      case 0x2003: temp="Falscher NCCI";break;
      case 0x2004: temp="Falscher Typ";break;

      case 0x3101: temp="B-Channel falsch kodiert";break;
      case 0x3102: temp="Info-Mask falsch kodiert";break;
      case 0x3103: temp="Serviced-EAZ-Mask falsch kodiert";break;
      case 0x3104: temp="Serviced-SI-Mask falsch kodiert";break;
      case 0x3105: temp="B2-Protocol falsch";break;
      case 0x3106: temp="DLPD falsch";break;
      case 0x3107: temp="B3-Protocol falsch";break;
      case 0x3108: temp="NCPD falsch";break;
      case 0x3109: temp="NCPI falsch";break;
      case 0x310A: temp="Flags falsch kodiert";break;

      case 0x3201: temp="Allgemeiner Fehler beim Controller";break;
      case 0x3202: temp="Konflikt zwischen Anmeldungen";break;
      case 0x3203: temp="Funktion wird nicht untersttzt";break;
      case 0x3204: temp="PLCI nicht aktiv";break;
      case 0x3205: temp="NCCI nicht aktiv";break;
      case 0x3206: temp="B2-Protokoll wird nicht untersttzt";break;
      case 0x3207: temp="Umschalten des B2-Protokolls in diesem Zustand nicht mglich";break;
      case 0x3208: temp="B3-Protokoll wird nicht untersttzt";break;
      case 0x3209: temp="Umschalten des B3-Protokolls in diesem Zustand nicht mglich";break;
      case 0x320A: temp="Verwendete Parameter im DLPD werden nicht untersttzt";break;
      case 0x320B: temp="Verwendete Parameter im NCPD werden nicht untersttzt";break;
      case 0x320C: temp="Verwendete Parameter im MCPI werden nicht untersttzt";break;
      case 0x320D: temp="Nicht untersttzte Data-Length";break;

      case 0x3301: temp="Fehler beim Aufbau D-Kanal Ebene 1";break;
      case 0x3302: temp="Fehler beim Aufbau D-Kanal Ebene 2";break;
      case 0x3303: temp="Fehler beim Aufbau B-Kanal Ebene 1";break;
      case 0x3304: temp="Fehler beim Aufbau B-Kanal Ebene 2";break;
      case 0x3305: temp="Abbruch D-Kanal Ebene 1";break;
      case 0x3306: temp="Abbruch D-Kanal Ebene 2";break;
      case 0x3307: temp="Abbruch D-Kanal Ebene 3";break;
      case 0x3308: temp="Abbruch B-Kanal Ebene 1";break;
      case 0x3309: temp="Abbruch B-Kanal Ebene 2";break;
      case 0x330A: temp="Abbruch B-Kanal Ebene 3";break;
      case 0x330C: temp="Reestablish B-Kanal Ebene 3";break;

      case 0x348a: temp="No channel available"; break;
      case 0x34a0: temp="Outgoing calls barred"; break;
      case 0x34a1: temp="User access busy"; break;
      case 0x34a3: temp="Nonexistent CUG"; break;
      case 0x34b5: temp="Destination not obtainable"; break;
      case 0x34b8: temp="Number changed"; break;
      case 0x34b9: temp="Out of order"; break;
      case 0x34ba: temp="No user responding"; break;
      case 0x34bb: temp="User busy"; break;
      case 0x34bd: temp="Incoming calls barred"; break;
      case 0x34be: temp="Call rejected"; break;
      case 0x34d9: temp="Network congestion"; break;
      case 0x34da: temp="Remote user initiated"; break;
      case 0x34f0: temp="Local procedure error"; break;

      case 0x4101: temp="Verlust der Frame-Synchronisation";

      default:
         return buff;
   }

   strcat (buff, temp);
#endif
   return buff;
}


char *api_getinfos (int number)
{
   union REGS r;
   struct SREGS s;

   mysegread (&s);
   r.h.ah = 0xf0+(byte) number;
   r.x.bx = FP_OFF (buff);
   s.es = FP_SEG (buff);
   myint86x (capi_int, &r, &r, &s);

   if (!buff[0])
      strcpy (buff, "-none-");
   else if (strlen (buff) > 40)
      strcpy (buff+37, "...");

   return buff;
}


void api_reregister (void)
{
   api_set_signal (0L);
   api_release ();
   smsgnr = 0;
   api_register ();
   api_set_signal ((void _FAR *) capi_event_entry);
   capi_listen_req ();
}


/* --- eof api.c */
